import logging
from datetime import timedelta
from threading import Lock
from typing import Optional, List

from contek_tusk.batching_config import BatchingConfig
from contek_tusk.entry_row import EntryRow
from contek_tusk.metric_data import MetricData
from contek_tusk.metric_formatter import MetricFormatter
from contek_tusk.table import Table

logger = logging.getLogger(__name__)


class MetricBatch:

    def __init__(
        self,
        table: Table,
        batching_config: BatchingConfig,
    ) -> None:
        self._table = table
        self._batching_config = batching_config
        self._rows: List[EntryRow] = []
        self._drop_count = 0
        self._lock = Lock()

    def is_immediate(self) -> bool:
        return self._batching_config.is_immediate()

    def get_period(self) -> timedelta:
        return self._batching_config.get_period()

    def add(self, entry_row: EntryRow) -> None:
        max_size = self._batching_config.get_max_size()
        self._lock.acquire()
        try:
            if 0 < max_size <= len(self._rows):
                self._drop_count = self._drop_count + 1
            else:
                self._rows.append(entry_row)
        finally:
            self._lock.release()

    def export(self, formatter: MetricFormatter) -> Optional[MetricData]:
        self._lock.acquire()
        try:
            data = formatter.format(self._rows)
            self._rows.clear()
            if self._drop_count > 0:
                logger.error(
                    f"Dropped {self._drop_count} rows inserting to table \"{self._table.get_full_name()}\"",
                )
                self._drop_count = 0
            return data
        finally:
            self._lock.release()



from logging import INFO, DEBUG, NOTSET, WARNING, ERROR, CRITICAL


class LevelMapper:

    def __call__(self, level: int) -> int:
        if level is NOTSET:
            return 0
        if level is CRITICAL:
            return 1
        if level is ERROR:
            return 2
        if level is WARNING:
            return 3
        if level is INFO:
            return 4
        if level is DEBUG:
            return 5
        return -1



from typing import Optional, Dict

from contek_tusk import app_name_resolver
from contek_tusk.metric import Metric
from contek_tusk.metric_client import MetricClient, DEFAULT_CLIENT

_clients: Dict[str, MetricClient] = {}


def init(
    host: str,
    user: str,
    password: str,
    client_name: str = DEFAULT_CLIENT,
    **kwargs: object,
) -> None:
    client = MetricClient.create(host, user, password, **kwargs)
    set_client(client, client_name)


def get_client(client_name: str = DEFAULT_CLIENT) -> Optional[MetricClient]:
    global _clients
    return _clients.get(client_name)


def set_client(
    client: MetricClient,
    client_name: str = DEFAULT_CLIENT,
) -> None:
    global _clients
    _clients[client_name] = client


metric = Metric.metric
set_app_name = app_name_resolver.set_app_name



from __future__ import annotations

import logging
from datetime import datetime, timedelta
from threading import RLock, Timer
from typing import Dict, Union

from contek_tusk.heartbeat.tuck_heartbeat_config import TuskHeartbeatConfig
from contek_tusk.metric import Metric
from contek_tusk.metric_client import DEFAULT_CLIENT

logger = logging.getLogger(__name__)


class TuskHeartbeatClient:
    _MAIN_TASK = 'main'

    def __init__(self, metric: Metric, config: TuskHeartbeatConfig) -> None:
        self._metric = metric
        self._config = config
        self._started: bool = False
        self._counts: Dict[str, int] = {}
        self._lock = RLock()

    @classmethod
    def create(
        cls,
        config: TuskHeartbeatConfig,
        client_name: str = DEFAULT_CLIENT,
    ) -> TuskHeartbeatClient:
        metric = Metric.metric(
            config.get_table(),
            config.get_batching_config(),
            client_name,
        )
        return cls(metric, config)

    def start(self):
        self._lock.acquire()
        try:
            if self._started:
                return
            self._send_main_beat_and_reschedule()
            self._started = True
        finally:
            self._lock.release()

    def beat(self, task: str, heartbeat_period: Union[timedelta, int]):
        if task == TuskHeartbeatClient._MAIN_TASK:
            raise ValueError(task)
        if type(heartbeat_period) is int:
            heartbeat_period = timedelta(seconds=heartbeat_period)
        self._send_beat(task, heartbeat_period)

    def _send_main_beat_and_reschedule(self):
        self._send_beat(
            TuskHeartbeatClient._MAIN_TASK,
            self._config.get_heartbeat_period(),
        )
        self._reschedule()

    def _reschedule(self):
        timer = Timer(
            self._config.get_heartbeat_period().total_seconds(),
            self._send_main_beat_and_reschedule,
        )
        timer.daemon = True
        timer.start()

    def _send_beat(self, task: str, heartbeat_period: timedelta):
        config = self._config

        app_name = config.get_app_name()
        now = datetime.utcnow()
        timeout_period = config.get_timeout()
        batching_period = config.get_batching_config().get_period()
        expiry = now + heartbeat_period + timeout_period + batching_period

        key_values: Dict[str, any] = {
            config.get_app_column(): app_name,
            config.get_task_column(): task,
            config.get_sequence_column(): self._next_sequence(task),
            config.get_expiry_column(): expiry,
        }
        self._metric.write(key_values)

    def _next_sequence(self, task: str) -> int:
        self._lock.acquire()
        try:
            count = self._counts.get(task)
            if count is None:
                count = 0
            self._counts[task] = count + 1
            return count
        finally:
            self._lock.release()



import os
from typing import Optional


def get_env_tag(column_name: str) -> Optional[str]:
    env_value = os.getenv(column_name)
    if env_value is None:
        env_value = os.getenv(column_name.upper())
    if env_value is None:
        env_value = os.getenv(column_name.lower())
    return env_value



from __future__ import annotations

import atexit
from datetime import datetime
from threading import Timer, RLock
from typing import Optional, Union, Dict

import contek_tusk as tusk
from contek_tusk.batching_config import BatchingConfig
from contek_tusk.entry_input_normalizer import EntryInputNormalizer
from contek_tusk.entry_row import EntryRow
from contek_tusk.env_tags_cache import EnvTagsCache
from contek_tusk.metric_batch import MetricBatch
from contek_tusk.metric_client import DEFAULT_CLIENT
from contek_tusk.metric_formatter import MetricFormatter
from contek_tusk.schema_provider import SchemaProvider
from contek_tusk.table import Table
from contek_tusk.time_column_cache import TimeColumnCache


class Metric:
    DEFAULT_BATCHING_CONFIG = BatchingConfig.default()

    def __init__(
        self,
        table: Table,
        time_column_cache: TimeColumnCache,
        env_tags_cache: EnvTagsCache,
        entry_input_normalizer: EntryInputNormalizer,
        metric_formatter: MetricFormatter,
        schema_provider: SchemaProvider,
        batching_config: BatchingConfig,
        client_name: str,
    ) -> None:
        self._table = table
        self._time_column_cache = time_column_cache
        self._env_tags_cache = env_tags_cache
        self._entry_input_normalizer = entry_input_normalizer
        self._metric_formatter = metric_formatter
        self._schema_provider = schema_provider
        self._batch = MetricBatch(table, batching_config)
        self._client_name = client_name
        self._lock: RLock = RLock()
        self._timer: Optional[Timer] = None
        atexit.register(self._flush)

    @classmethod
    def metric(
        cls,
        table: Union[Table, str],
        batching_config: BatchingConfig = DEFAULT_BATCHING_CONFIG,
        client_name: str = DEFAULT_CLIENT,
    ) -> Metric:
        if type(table) is str:
            table = Table.from_str(table)
        schema_provider = SchemaProvider(table, client_name)
        return cls(
            table,
            TimeColumnCache(schema_provider, table.get_time_column()),
            EnvTagsCache(schema_provider),
            EntryInputNormalizer(schema_provider),
            MetricFormatter(table, schema_provider),
            schema_provider,
            batching_config,
            client_name,
        )

    def write(self, key_values: Dict[str, any]) -> None:
        time_column = self._time_column_cache.get()
        if time_column is not None:
            now = datetime.utcnow()
            key_values[time_column] = now

        env_tags = self._env_tags_cache.get()
        if env_tags is not None:
            for (key, value) in env_tags.items():
                if key not in key_values:
                    key_values[key] = value

        entry_row = self._entry_input_normalizer.normalize(key_values)
        if entry_row is None:
            return

        self._accept(entry_row)

    def get_schema_provider(self) -> SchemaProvider:
        return self._schema_provider

    def _accept(self, entry_row: EntryRow) -> None:
        self._batch.add(entry_row)
        self._schedule_if_idle()

    def _schedule_if_idle(self) -> None:
        if self._batch.is_immediate():
            self._flush()
            return
        self._lock.acquire()
        try:
            timer = self._timer
            if timer is not None and timer.is_alive():
                return
            self._schedule()
        finally:
            self._lock.release()

    def _flush_and_schedule(self) -> None:
        updated = self._flush()
        if not updated:
            return
        self._schedule()

    def _schedule(self) -> None:
        self._lock.acquire()
        try:
            timer = Timer(
                self._batch.get_period().total_seconds(),
                self._flush_and_schedule,
            )
            timer.daemon = True
            timer.start()
            self._timer = timer
        finally:
            self._lock.release()

    def _flush(self) -> bool:
        client = tusk.get_client(self._client_name)
        if client is None:
            return False

        data = self._batch.export(self._metric_formatter)
        if data is None:
            return False

        client.write(data)
        return True



from datetime import timedelta

from contek_tusk.app_name_resolver import APP_NAME_KEY
from contek_tusk.batching_config import BatchingConfig
from contek_tusk.table import Table


class TuskHeartbeatConfig:
    DEFAULT_BATCHING_CONFIG = BatchingConfig.unlimited(timedelta(seconds=5))
    DEFAULT_APP_COLUMN = APP_NAME_KEY
    DEFAULT_TASK_COLUMN = 'task'
    DEFAULT_SEQUENCE_COLUMN = 'sequence'
    DEFAULT_EXPIRY_COLUMN = 'expiry'

    def __init__(
        self,
        app_name: str,
        table: Table,
        heartbeat_period: timedelta,
        timeout: timedelta,
        batching_config: BatchingConfig = DEFAULT_BATCHING_CONFIG,
        app_column: str = DEFAULT_APP_COLUMN,
        task_column: str = DEFAULT_TASK_COLUMN,
        sequence_column: str = DEFAULT_SEQUENCE_COLUMN,
        expiry_column: str = DEFAULT_EXPIRY_COLUMN,
    ) -> None:
        self._app_name = app_name
        self._table = table
        self._heartbeat_period = heartbeat_period
        self._timeout = timeout
        self._batching_config = batching_config
        self._app_column = app_column
        self._task_column = task_column
        self._sequence_column = sequence_column
        self._expiry_column = expiry_column

    def get_app_name(self) -> str:
        return self._app_name

    def get_table(self) -> Table:
        return self._table

    def get_heartbeat_period(self) -> timedelta:
        return self._heartbeat_period

    def get_timeout(self) -> timedelta:
        return self._timeout

    def get_batching_config(self) -> BatchingConfig:
        return self._batching_config

    def get_app_column(self) -> str:
        return self._app_column

    def get_task_column(self) -> str:
        return self._task_column

    def get_sequence_column(self) -> str:
        return self._sequence_column

    def get_expiry_column(self) -> str:
        return self._expiry_column



from typing import Dict


class CountDetails:

    def __init__(self, tags: Dict[str, str]) -> None:
        self._tags = tags
        self._count = 0

    def get_tags(self) -> Dict[str, str]:
        return self._tags

    def increment_unsafe(self, n: int) -> None:
        self._count = self._count + n

    def get_count_unsafe(self) -> int:
        return self._count
